import asyncio
import aiohttp
import aiobotocore
from urllib import parse


class S3ClientPool:
    def __init__(self, region_name, size):
        self.session = aiobotocore.get_session()
        self.size = size
        self.semaphore = asyncio.Semaphore(size)
        self.clients = [
            self.session.create_client('s3', region_name=region_name)
            for i in range(size)
        ]

    async def acquire(self):
        await self.semaphore.acquire()
        return self.clients.pop(0)

    def recycle(self, client):
        self.clients.append(client)
        self.semaphore.release()

    def release(self):
        for client in self.clients:
            client.close()


class S3Uploader:
    def __init__(self, bucket, region_name, prefix=None, concurrency=3):
        self.bucket = bucket
        self.region_name = region_name
        self.prefix = prefix
        self.concurrency = concurrency
        self.pool = S3ClientPool(region_name, concurrency)

    async def s3_upload(self, info):
        session = aiohttp.ClientSession()
        async with session:
            async with session.get(info['url']) as response:
                body = await response.read()
        client = await self.pool.acquire()
        try:
            suffix = parse.urlparse(info['url']).path.split('.')[-1]
            if suffix not in ['jpg', 'png', 'jpeg']:
                suffix = 'jpg'
            key = info['hash'] + '.' + suffix
            if self.prefix is not None:
                key = self.prefix + '/' + key
            print('upload image from {} to {} {}'.format(
                info['url'], key, response.content_type))
            resp = await client.put_object(Bucket=self.bucket,
                                           Key=key,
                                           Body=body,
                                           ContentType=response.content_type)
            print(resp)
        finally:
            self.pool.recycle(client)

    def upload_all(self, infos):
        loop = asyncio.get_event_loop()
        coro = batch_ops(infos, self.concurrency, self.s3_upload)
        loop.run_until_complete(coro)

    def __del__(self):
        self.pool.release()


async def batch_ops(obj_list, concurrency, coro_func, timeout=None):
    obj_iter = iter(obj_list)
    result = []
    pending = []
    while True:
        try:
            obj = next(obj_iter)
        except StopIteration:
            pass
        else:
            job = coro_func(obj)
            pending.append(job)
            if len(pending) < concurrency:
                continue
        if not pending:
            break
        done, pending_set = await asyncio.wait(
            pending, timeout=timeout, return_when=asyncio.FIRST_COMPLETED)
        pending = list(pending_set)
        result.append(done.pop())
    return result


def main():
    s3_uploader = S3Uploader('test-gyb', 'us-west-1')
    s3_uploader.upload_all([
        'http://i.imgur.com/l0sYs8S.jpg',
        'http://i.imgur.com/hBuSQA5.png',
        'http://i.imgur.com/Z65buhC.jpg',
        'http://i.imgur.com/wGlA8A4.jpg',
    ])


if __name__ == '__main__':
    main()



import time
import json
import asyncio
import logging
import aiobotocore
from .pool import Pool
from .worker import BatchWorker
logger = logging.getLogger(__name__)


class Firehose:
    service_name = 'firehose'

    def __init__(self, stream_name, region_name,
                 encode_func=None, delimiter=b'\n',
                 aggr_num=100, flush_interval=60,
                 **pool_kwargs):
        self.session = aiobotocore.get_session()
        self.stream_name = stream_name
        self.region_name = region_name
        self.encode_func = encode_func or (lambda o: json.dumps(o).encode())
        self.delimiter = delimiter
        self.jobs = set()
        self.pool = Pool(self.create_client, **pool_kwargs)
        self.worker = BatchWorker(
            self.handle, aggr_num=aggr_num, timeout=flush_interval
        )

    def __repr__(self):
        return '<{}: stream={}, region={}, worker={!r}, pool={!r}>'.format(
                self.__class__.__name__, self.stream_name,
                self.region_name, self.worker, self.pool)

    async def put(self, obj):
        return await self.worker.put(obj)

    def create_client(self):
        return self.session.create_client(
            'firehose', region_name=self.region_name
        )

    async def stop(self):
        timestamp = time.time()
        await self.worker.stop()
        if self.jobs:
            await asyncio.wait(self.jobs)
        await self.pool.close()
        cost = time.time() - timestamp
        logger.info('{0!r} stopped in {1:.1f} seconds'.format(self, cost))

    def _encode(self, obj_list):
        encoded = [self.encode_func(obj) for obj in obj_list]
        encoded.append(b'')
        return self.delimiter.join(encoded)

    async def handle(self, obj_list):
        data = self._encode(obj_list)
        client = await self.pool.acquire()
        fut = asyncio.ensure_future(
            self.pool.auto_release(client, client.put_record(
                DeliveryStreamName=self.stream_name,
                Record={'Data': data}
            ))
        )
        self.jobs.add(fut)
        fut.add_done_callback(self.jobs.remove)



import aiobotocore
from pynamodb.constants import (
    RETURN_CONSUMED_CAPACITY_VALUES, RETURN_ITEM_COLL_METRICS_VALUES, COMPARISON_OPERATOR_VALUES,
    RETURN_ITEM_COLL_METRICS, RETURN_CONSUMED_CAPACITY, RETURN_VALUES_VALUES, ATTR_UPDATE_ACTIONS,
    COMPARISON_OPERATOR, EXCLUSIVE_START_KEY, SCAN_INDEX_FORWARD, SCAN_FILTER_VALUES, ATTR_DEFINITIONS,
    BATCH_WRITE_ITEM, CONSISTENT_READ, ATTR_VALUE_LIST, DESCRIBE_TABLE, KEY_CONDITIONS,
    BATCH_GET_ITEM, DELETE_REQUEST, SELECT_VALUES, RETURN_VALUES, REQUEST_ITEMS, ATTR_UPDATES,
    ATTRS_TO_GET, SERVICE_NAME, DELETE_ITEM, PUT_REQUEST, UPDATE_ITEM, SCAN_FILTER, TABLE_NAME,
    INDEX_NAME, KEY_SCHEMA, ATTR_NAME, ATTR_TYPE, TABLE_KEY, EXPECTED, KEY_TYPE, GET_ITEM, UPDATE,
    PUT_ITEM, SELECT, ACTION, EXISTS, VALUE, LIMIT, QUERY, SCAN, ITEM, LOCAL_SECONDARY_INDEXES,
    KEYS, KEY, EQ, SEGMENT, TOTAL_SEGMENTS, CREATE_TABLE, PROVISIONED_THROUGHPUT, READ_CAPACITY_UNITS,
    WRITE_CAPACITY_UNITS, GLOBAL_SECONDARY_INDEXES, PROJECTION, EXCLUSIVE_START_TABLE_NAME, TOTAL,
    DELETE_TABLE, UPDATE_TABLE, LIST_TABLES, GLOBAL_SECONDARY_INDEX_UPDATES,
    CONSUMED_CAPACITY, CAPACITY_UNITS, QUERY_FILTER, QUERY_FILTER_VALUES, CONDITIONAL_OPERATOR,
    CONDITIONAL_OPERATORS, NULL, NOT_NULL, SHORT_ATTR_TYPES, DELETE,
    ITEMS, DEFAULT_ENCODING, BINARY_SHORT, BINARY_SET_SHORT, LAST_EVALUATED_KEY, RESPONSES, UNPROCESSED_KEYS,
    UNPROCESSED_ITEMS, STREAM_SPECIFICATION, STREAM_VIEW_TYPE, STREAM_ENABLED)
BOTOCORE_EXCEPTIONS = (BotoCoreError, ClientError)


class Connection:
    service_name = 'dynamodb'

    def __init__(self, region_name, host=None):
        self.session = aiobotocore.get_session()
        self._client = None
        self._tables = {}

    def create_client(self):
        return self.session.create_client(
            self.service_name, region_name=self.region_name
        )

    @property
    def client(self):
        if self._client is None:
            self._client = self.create_client()
        return self._client

    async def get_item(self,
                       table_name,
                       hash_key,
                       range_key=None,
                       consistent_read=False,
                       attributes_to_get=None):
        operation_kwargs = {}
        if attributes_to_get is not None:
            operation_kwargs[ATTRS_TO_GET] = attributes_to_get
        operation_kwargs[CONSISTENT_READ] = consistent_read
        operation_kwargs[TABLE_NAME] = table_name
        operation_kwargs.update(
            self.get_identifier_map(table_name, hash_key, range_key)
        )
        return await self.client.get_item(
        try:
            return self.dispatch(GET_ITEM, operation_kwargs)
        except BOTOCORE_EXCEPTIONS as e:
            raise GetError("Failed to get item: {0}".format(e), e)

    def get_identifier_map(self, table_name, hash_key, range_key=None, key=KEY):
        """
        Builds the identifier map that is common to several operations
        """
        tbl = self.get_meta_table(table_name)
        if tbl is None:
            raise TableError("No such table {0}".format(table_name))
        return tbl.get_identifier_map(hash_key, range_key=range_key, key=key)

    def get_meta_table(self, table_name, refresh=False):
        """
        Returns a MetaTable
        """
        if table_name not in self._tables or refresh:
            operation_kwargs = {
                TABLE_NAME: table_name
            }
            try:
                data = self.dispatch(DESCRIBE_TABLE, operation_kwargs)
                self._tables[table_name] = MetaTable(data.get(TABLE_KEY))
            except BotoCoreError as e:
                raise TableError("Unable to describe table: {0}".format(e), e)
            except ClientError as e:
                if 'ResourceNotFound' in e.response['Error']['Code']:
                    raise TableDoesNotExist(e.response['Error']['Message'])
                else:
                    raise
        return self._tables[table_name]

    async def dispatch(self, operation_name, operation_kwargs):
        if



import math
import time
from contextlib import contextmanager


class GaugeValue:
    def __init__(self):
        self.val = 0
        self.val_min = float('inf')
        self.val_max = float('-inf')

    def add(self, val):
        new_val = self.val + val
        self.set(new_val)

    def set(self, val):
        self.val_min = min(val, self.val_min)
        self.val_max = max(val, self.val_max)
        self.val = val

    def reset(self):
        self.val_max = self.val_min = self.val


def get_stats(lst):
    if not lst:
        return {}
    lst.sort()
    total = sum(lst)
    length = len(lst)
    avg = total / length
    p95 = math.ceil(length * 95 / 100) - 1
    p5 = math.ceil(length * 5 / 100) - 1
    return {
        'sum': total,
        'avg': avg,
        'max_p95': lst[p95],
        'min_p95': lst[p5]
    }


class LocalStats:
    def __init__(self, default=None):
        self.default = default or {}
        self.metrics = {}
        self.functions = {}

    def incr(self, stat, count=1, rate=1):
        if stat not in self.metrics:
            self.metrics[stat] = 0
        self.metrics[stat] += count / rate

    def decr(self, stat, count=1, rate=1):
        if stat not in self.metrics:
            self.metrics[stat] = 0
        self.metrics[stat] -= count / rate

    def gauge(self, stat, value, delta=False):
        if stat not in self.metrics:
            self.metrics[stat] = GaugeValue()
        if delta:
            self.metrics[stat].add(value)
        else:
            self.metrics[stat].set(value)

    def timing(self, stat, delta):
        if stat not in self.metrics:
            self.metrics[stat] = []
        self.metrics[stat].append(delta)

    @contextmanager
    def timer(self, stat):
        start = time.time()
        yield
        end = time.time()
        self.timing(stat, int((end-start) * 1000))

    def unique(self, stat, value):
        if stat not in self.metrics:
            self.metrics[stat] = set()
        self.metrics[stat].add(value)

    def before_flush(self, stat, func):
        self.functions[stat] = func

    def _flush(self):
        result = {}
        for stat, func in self.functions.items():
            self.gauge(stat, func())
        for k, v in self.metrics.items():
            if isinstance(v, GaugeValue):
                result[k] = v.val
                result['%s_max' % k] = v.val_max
                result['%s_min' % k] = v.val_min
                v.reset()
            elif isinstance(v, (int, float)):
                result[k] = v
                self.metrics[k] = 0
            elif isinstance(v, set):
                result[k] = len(v)
                v.clear()
            elif isinstance(v, list):
                for postfix, val in get_stats(v).items():
                    result['%s_%s' % (k, postfix)] = val
                v.clear()
        return result

    def flush(self):
        data = self._flush()
        data.update(self.default)
        return data
